import numpy as np
import os
import pandas as pd
import random
import torch.utils.data
import torch
from myutils import *
from helper_code import *
from skimage.transform import resize
from scipy.signal import medfilt
from scipy.io import loadmat
from scipy.signal import hilbert


# Single position 1D signal output
class MurmurSingleChannelDataset1D(torch.utils.data.Dataset):
    def __init__(self, files_pd, resize_scale=32, signal_length=8192, decompose=True, dataset=2022, resample=True,
                 ori_mask=True):
        super(MurmurSingleChannelDataset1D, self).__init__()
        self.files = files_pd['FileName'].values  # file absolute path
        self.labels = files_pd['Label'].values.astype(int)
        self.maskfiles = files_pd['MaskFile'].values
        self.resize_scale = resize_scale
        self.signal_length = signal_length
        self.decompose = decompose
        self.dataset = dataset
        self.resample = resample
        self.ori_mask = ori_mask
        label_type = set(self.labels)
        label_weight = {}
        label_count = {}
        for l in label_type:
            label_weight[l] = 1. / len(np.where(self.labels == l)[0])
            label_count[l] = len(np.where(self.labels == l)[0])
        # sampler weights
        weights = []
        for l in self.labels:
            weights.append(label_weight[l])
        print(label_count)
        num_labels = len(label_count)
        data_list = []
        for i in range(num_labels):
            idx = np.where(self.labels == i)[0]
            data_list.append(idx)
        self.data = data_list
        self.weights = np.array(weights)
        self.num_ = len(label_count) * np.max(list(label_count.values()))
        self.minor_class_count = int(np.min(list(label_count.values())) * len(label_count))

    def __len__(self):
        return len(self.files)

    def __getitem__(self, idx):
        if self.resample:
            cls = idx % len(self.data)
            idx = self.data[cls][(idx // len(self.data)) % len(self.data[cls])]
        file = self.files[idx]
        label = int(self.labels[idx])
        maskfile = self.maskfiles[idx]
        # output
        # Pre-processed one position PCG (5 channel = orignal + decomposed 4)  + label
        out_sig, out_mask = self.load(file, maskfile, label)
        # check if len are correct
        new_out_sig = []
        for sig in out_sig:
            if len(sig) == self.signal_length:
                new_out_sig.append(sig)
            else:
                if len(sig) < self.signal_length:
                    new_out_sig.append(np.pad(sig, (0, self.signal_length - len(sig))))
                else:
                    new_out_sig.append(sig[:self.signal_length])
        # check mask length
        if len(out_mask) == self.signal_length:
            new_out_mask = out_mask.copy()
        else:
            if len(sig) < self.signal_length:
                new_out_mask = np.pad(out_mask, (0, self.signal_length - len(sig)))
            else:
                new_out_mask = out_mask[:self.signal_length]

        # resize
        new_out_sig = [decimate(sig, self.resize_scale, ftype='fir') for sig in new_out_sig]
        new_out_mask = np.round(decimate(new_out_mask, self.resize_scale, ftype='fir')).astype(int)
        new_out_mask4 = np.zeros((5, len(new_out_mask)))
        for i in range(1, 5):
            new_out_mask4[i, np.where(new_out_mask == i)[0]] = 1
        return np.array(new_out_sig), new_out_mask, new_out_mask4, int(label)

    def load(self, filepath, maskfilepath, label):
        record, fre = load_wav_file(filepath)
        if self.dataset == 2022:
            features, sm, dm = get_header_features(filepath)
            mask, _ = read_mask2022(maskfilepath, len(record), sm, dm, label, ori=self.ori_mask)
        else:
            mask, _ = read_mask2016(maskfilepath, len(record))
        # assert (len(record) == len(mask))
        # get rid of no annotation part
        annot_s, annot_e = 0, len(mask)
        for i in range(len(mask)):
            if mask[i] != 0:
                annot_s = i
                break
        for i in range(len(mask) - 1, 0, -1):
            if mask[i] != 0:
                annot_e = i
                break
        record = record[annot_s:annot_e]
        mask = mask[annot_s:annot_e]
        all_len = len(record)
        # random cut signal
        if all_len > self.signal_length:
            s = np.random.randint(0, all_len - self.signal_length + 1, 1)[0]
            record = record[s:s + self.signal_length]
            mask = mask[s:s + self.signal_length]
        else:
            record = np.pad(record, (0, self.signal_length - all_len))
            mask = np.pad(mask, (0, self.signal_length - all_len))
        # downsample
        # drecordings = downsample(record, factor=int(fre/self.frequency))
        # dmask = np.round(downsample(mask, factor=int(fre/self.frequency))).astype(int)
        # filtering
        frecordings = filtering(record, low=25, high=400, order=4, sampling_rate=fre)
        # remove spikes
        srecordings = remove_spikes(frecordings, terminal=500, sampling=fre)
        # normalize
        nrecordings = normalize_signal(srecordings)
        # decomposition
        if self.decompose:
            decomp_records = decompose_into_n_signals(nrecordings)
            decomp_records = [np.abs(hilbert(r)) for r in decomp_records]
            decomp_records.append(np.abs(hilbert(nrecordings)))
            decomp_records = np.array(decomp_records)
        else:
            decomp_records = np.array(np.abs(hilbert(nrecordings)))
        return decomp_records, mask


class MurmurSingleChannelHeartBeatDataset2D(torch.utils.data.Dataset):
    def __init__(self, files_pd, img_size=64, hop_length=64, n_fft=512, dataset=2022, prob=0.7, all_cycles=False):
        super(MurmurSingleChannelHeartBeatDataset2D, self).__init__()
        self.files = files_pd['FileName'].values  # file absolute path
        self.labels = files_pd['Label'].values.astype(int)
        self.maskfiles = files_pd['MaskFile'].values
        self.img_size = img_size
        self.hop_length = hop_length
        self.n_fft = n_fft
        self.dataset = dataset
        self.prob = prob
        self.all_cycles = all_cycles
        label_type = set(self.labels)
        label_weight = {}
        label_count = {}
        for l in label_type:
            label_weight[l] = 1. / len(np.where(self.labels == l)[0])
            label_count[l] = len(np.where(self.labels == l)[0])
        # sampler weights
        weights = []
        for l in self.labels:
            weights.append(label_weight[l])
        print(label_count)
        num_labels = len(label_count)
        data_list = []
        for i in range(num_labels):
            idx = np.where(self.labels==i)[0]
            data_list.append(idx)
        self.data = data_list
        self.weights = np.array(weights)
        self.num_ = len(label_count) * np.max(list(label_count.values()))
        self.minor_class_count = int(np.min(list(label_count.values())) * len(label_count))

    def __len__(self):
        return len(self.files)

    def __getitem__(self, idx):
        cls = idx % len(self.data)
        new_idx = self.data[cls][(idx//len(self.data)) % len(self.data[cls])]
        file = self.files[new_idx]
        label = int(self.labels[new_idx])
        maskf = self.maskfiles[new_idx]
        spec_img = self.load(file, maskf)
        return spec_img, int(label)

    def load(self, filepath, maskfilepath):
        record, fre = load_wav_file(filepath)
        if self.dataset == 2022:
            mask, cycle_list = read_mask2022(maskfilepath, len(record))
        else:
            mask, cycle_list = read_mask2016(maskfilepath, len(record))
        # assert (len(record) == len(mask))
        # get rid of no annotation part
        annot_s, annot_e = 0, len(mask)
        for i in range(len(mask)):
            if mask[i] != 0:
                annot_s = i
                break
        for i in range(len(mask) - 1, 0, -1):
            if mask[i] != 0:
                annot_e = i
                break
        record = record[annot_s:annot_e]
        mask = mask[annot_s:annot_e]
        # filtering
        record = filtering(record, low=25, high=500, order=4, sampling_rate=fre)
        # remove spikes
        record = remove_spikes(record, terminal=500, sampling=fre)
        # normalize
        record = normalize_signal(record)
        cycle_list = cycle_list - annot_s
        # choose one cycle + augmentation
        len_record = len(record)
        if not self.all_cycles:
            if len(cycle_list)==0:
                se = (fre * 60./80)
                s = np.random.randint(0, len_record - se - 1, 1)[0]
                e = s + se
            else:
                k = np.random.randint(0, len(cycle_list), 1)[0]
                s, e = cycle_list[k]
            len_cycle = int(e - s)
            # add some randomness
            if np.random.rand(1)[0] > self.prob:
                slide = (np.random.rand(1)[0] - 0.5) * 2 * 0.2 * len_cycle
                s = max(0, s + slide)
            if np.random.rand(1)[0] > self.prob:
                slide = (np.random.rand(1)[0] - 0.5) * 2 * 0.2 * len_cycle
                e = min(len_record, e + slide)
            one_cycle = record[int(s):int(e)]
            img = create_2d_sepctrogram(one_cycle, hop_length=self.hop_length, n_fft=self.n_fft,
                                        sample_rate=fre)
            # resize image to [img_size, img_size]
            img = resize(img.astype(float), (self.img_size, self.img_size), 1)[::-1]
            img = (img - np.min(img)) / (np.max(img) - np.min(img) + 1e-5)
            return np.expand_dims(img, axis=0)
        else:
            all_imgs = []
            for i in range(len(cycle_list)):
                s, e = cycle_list[i]
                len_cycle = int(e - s)
                # add some randomness
                if np.random.rand(1)[0] > self.prob:
                    slide = (np.random.rand(1)[0] - 0.5) * 2 * 0.2 * len_cycle
                    s = max(0, s + slide)
                if np.random.rand(1)[0] > self.prob:
                    slide = (np.random.rand(1)[0] - 0.5) * 2 * 0.2 * len_cycle
                    e = min(len_record, e + slide)
                one_cycle = record[int(s):int(e)]
                img = create_2d_sepctrogram(one_cycle, hop_length=self.hop_length, n_fft=self.n_fft,
                                            sample_rate=fre)
                # resize image to [img_size, img_size]
                img = resize(img.astype(float), (self.img_size, self.img_size), 1)[::-1]
                img = (img - np.min(img)) / (np.max(img) - np.min(img) + 1e-5)
                all_imgs.append(img)
            return np.array(all_imgs)



# signal signal 2D input
class MurmurSingleChannelDataset2D(torch.utils.data.Dataset):
    def __init__(self, files_pd, frequency=4000, img_size=128, hop_length=64, n_fft=512, dataset=2022):
        super(MurmurSingleChannelDataset2D, self).__init__()
        self.files = files_pd['FileName'].values  # file absolute path
        self.labels = files_pd['Label'].values
        self.maskfiles = files_pd['MaskFile'].values
        self.frequency = frequency
        self.img_size = img_size
        self.hop_length = hop_length
        self.n_fft = n_fft
        self.dataset = dataset
        label_type = set(self.labels)
        label_weight = {}
        label_count = {}
        for l in label_type:
            label_weight[l] = 1. / len(np.where(self.labels == l)[0])
            label_count[l] = len(np.where(self.labels == l)[0])
        # sampler weights
        weights = []
        for l in self.labels:
            weights.append(label_weight[l])
        self.weights = np.array(weights)
        self.minor_class_count = int(np.min(list(label_count.values())) * len(label_count))

    def __len__(self):
        return len(self.files)

    def __getitem__(self, idx):
        file = self.files[idx]
        label = self.labels[idx]
        maskf = self.maskfiles[idx]
        signal, mask = self.load(file, maskf, label)
        return signal, mask, int(label)

    def load(self, filepath, maskfilepath, label):
        record, fre = load_wav_file(filepath)
        if self.dataset == 2022:
            features, sm, dm = get_header_features(filepath)
            mask = read_mask2022(maskfilepath, len(record), sm, dm, label, ori=False)
            mask = medfilt(mask, kernel_size=self.n_fft + 1)
        else:
            mask = read_mask2016(maskfilepath, len(record))
            mask = medfilt(mask, kernel_size=self.n_fft + 1)
        # assert (len(record)==len(mask))
        # get rid of no annotation part
        annot_s, annot_e = 0, len(mask)
        for i in range(len(mask)):
            if mask[i] != 0:
                annot_s = i
                break
        for i in range(len(mask) - 1, 0, -1):
            if mask[i] != 0:
                annot_e = i
                break
        record = record[annot_s:annot_e]
        mask = record[annot_s:annot_e] - 1
        # filtering
        frecordings = filtering(record, low=25, high=500, order=4, sampling_rate=self.frequency)
        # remove spikes
        srecordings = remove_spikes(frecordings, terminal=500, sampling=self.frequency)
        # normalize
        nrecordings = normalize_signal(srecordings)
        # cut signal into small piece
        leng = int(self.img_size * self.hop_length)
        start_idx = np.random.randint(0, len(nrecordings) - leng, 1)[0]
        signal_piece = nrecordings[start_idx: start_idx + leng]
        mask_piece = mask[start_idx: start_idx + leng]
        # turn into spectrogram
        img = create_2d_sepctrogram(signal_piece, hop_length=self.hop_length, n_fft=self.n_fft,
                                    sample_rate=self.frequency)[:, :self.img_size]
        seg = mask_piece[np.arange(0, len(signal_piece), self.hop_length)][1:self.img_size + 1]
        # resize
        img = resize(img.astype(float), (self.img_size, self.img_size), 1)
        # normalize to [0,1]
        img = (img - np.min(img)) / (np.max(img) - np.min(img))
        seg_out = np.zeros((4, len(seg)))
        for i in range(4):
            seg_out[i, np.where(seg == i)[0]] = 1
        return np.expand_dims(img, 0), seg_out


def read_mask2022(filename, len_signal, sm=None, dm=None, label=None, ori=True):
    df = pd.read_csv(filename, sep='\t', header=None)
    segs = df.values
    segs[:, :2] = np.round(segs[:, :2] * 4000).astype(int)
    seg_seq = []
    for i in range(len(segs)):
        s, e, flag = segs[i]
        s = int(s)
        e = int(e)
        if i == 0:
            if s != 0:
                s = 0
        if i == len(segs) - 1:
            if e != len_signal:
                e = len_signal
        seg_seq.append([flag] * (e - s))
    seg_seq = np.concatenate(seg_seq)
    s1_cycle = segs[np.where(segs[:, 2] == 1)[0]]
    cycle_list = []
    for i in range(len(s1_cycle) - 1):
        cycle_list.append([s1_cycle[i][0], s1_cycle[i + 1][0]])
    if ori:
        return seg_seq, np.array(cycle_list)
    else:
        # add murmur (2)  + silent (4)
        # read file header
        tmp = seg_seq.copy()
        idx2 = np.where(seg_seq==2)[0]
        idx4 = np.where(seg_seq==4)[0]
        tmp[idx2] = 4
        if label==0 and sm==1:
            tmp[idx2] = 2
        if label==0 and dm==1:
            tmp[idx4] = 2
        return tmp, np.array(cycle_list)


def get_header_features(filepath):
    prefix = '/'.join(filepath.split('/')[:-1])
    txt_file = prefix + '/' + filepath[:-4].split('/')[-1].split('_')[0] + '.txt'
    data = load_patient_data(txt_file)
    age_group = get_age(data)
    sex = get_sex(data)
    height = get_height(data)
    weight = get_weight(data)
    is_pregnant = get_pregnancy_status(data)
    sm = get_systolic_murmur(data)
    dm = get_diastolic_murmur(data)
    # age coding
    if compare_strings(age_group, 'Neonate'):
        age = 0.5
    elif compare_strings(age_group, 'Infant'):
        age = 6
    elif compare_strings(age_group, 'Child'):
        age = 12
    elif compare_strings(age_group, 'Adolescent'):
        age = 15
    elif compare_strings(age_group, 'Young Adult'):
        age = 20
    else:
        age = 0
    # sex. Use one-hot encoding.
    sex_features = np.zeros(2, dtype=int)
    if compare_strings(sex, 'Female'):
        sex_features[0] = 1
    elif compare_strings(sex, 'Male'):
        sex_features[1] = 1
    features = np.hstack(([age], sex_features, [height], [weight], [is_pregnant]))
    return features, sm, dm


def read_mask2016(filename, len_signal):
    df = loadmat(filename)
    ans = df['state_ans']
    ans_arr = []
    states = {
        'S1': 1,
        'systole': 2,
        'S2': 3,
        'diastole': 4
    }
    for i in range(len(ans)):
        s = int(ans[i][0])
        try:
            t = str(ans[i][1][0, 0][0])
        except:
            t = str(ans[i][1][0])
        if i == len(ans) - 1:
            e = len_signal
        else:
            e = int(ans[i + 1][0])
        if t not in states.keys():
            ans_arr.append([s, e, 0])
        else:
            ans_arr.append([s, e, states[t]])
    segs = np.array(ans_arr)
    seg_seq = []
    for i in range(len(segs)):
        s, e, flag = segs[i]
        s = int(s)
        e = int(e)
        seg_seq.append([flag] * (min(e, len_signal) - s))
    seg_seq.append([flag])
    seg_seq = np.concatenate(seg_seq)
    s1_cycle = segs[np.where(segs[:, 2] == 1)[0]]
    cycle_list = []
    for i in range(len(s1_cycle) - 1):
        cycle_list.append([s1_cycle[i][0], s1_cycle[i + 1][0]])
    return seg_seq, np.array(cycle_list)
