import torch.nn as nn
import torch.nn.functional as F

#==============================================================================
# 2D model
#==============================================================================
def dwise_conv(ch_in, stride=1):
    return (
        nn.Sequential(
            #depthwise
            nn.Conv2d(ch_in, ch_in, kernel_size=3, padding=1, stride=stride, groups=ch_in, bias=False),
            nn.BatchNorm2d(ch_in),
            nn.ReLU6(inplace=True),
        )
    )


def conv1x1(ch_in, ch_out):
    return (
        nn.Sequential(
            nn.Conv2d(ch_in, ch_out, kernel_size=1, padding=0, stride=1, bias=False),
            nn.BatchNorm2d(ch_out),
            nn.ReLU6(inplace=True)
        )
    )


def conv3x3(ch_in, ch_out, stride):
    return (
        nn.Sequential(
            nn.Conv2d(ch_in, ch_out, kernel_size=3, padding=1, stride=stride, bias=False),
            nn.BatchNorm2d(ch_out),
            nn.ReLU6(inplace=True)
        )
    )


class InvertedBlock(nn.Module):
    def __init__(self, ch_in, ch_out, expand_ratio, stride):
        super(InvertedBlock, self).__init__()

        self.stride = stride
        assert stride in [1,2]

        hidden_dim = ch_in * expand_ratio

        self.use_res_connect = self.stride==1 and ch_in==ch_out

        layers = []
        if expand_ratio != 1:
            layers.append(conv1x1(ch_in, hidden_dim))
        layers.extend([
            #dw
            dwise_conv(hidden_dim, stride=stride),
            #pw
            conv1x1(hidden_dim, ch_out)
        ])

        self.layers = nn.Sequential(*layers)

    def forward(self, x):
        if self.use_res_connect:
            return x + self.layers(x)
        else:
            return self.layers(x)


class MobileNetV2(nn.Module):
    def __init__(self, ch_in=3, n_classes=1000):
        super(MobileNetV2, self).__init__()

        self.configs=[
            # t, c, n, s
            [1, 16, 1, 1],
            [6, 24, 2, 2],
            [6, 32, 3, 2],
            [6, 64, 4, 2],
            [6, 96, 3, 1],
            [6, 160, 3, 2],
            [6, 320, 1, 1]
        ]

        self.stem_conv = conv3x3(ch_in, 32, stride=2)

        layers = []
        input_channel = 32
        for t, c, n, s in self.configs:
            for i in range(n):
                stride = s if i == 0 else 1
                layers.append(InvertedBlock(ch_in=input_channel, ch_out=c, expand_ratio=t, stride=stride))
                input_channel = c

        self.layers = nn.Sequential(*layers)

        self.last_conv = conv1x1(input_channel, 1280)

        self.classifier = nn.Sequential(
            nn.Dropout2d(0.2),
            nn.Linear(1280, 32),
            nn.Linear(32, n_classes)
        )
        self.avg_pool = nn.AdaptiveAvgPool2d(1)

    def forward(self, x):
        x = self.stem_conv(x)
        x = self.layers(x)
        x = self.last_conv(x)
        x = self.avg_pool(x).view(-1, 1280)
        x = self.classifier(x)
        return x


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(
            in_planes, planes, kernel_size=3, stride=stride, padding=1)#, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3,
                               stride=1, padding=1)#, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion*planes,
                          kernel_size=1, stride=stride),#, bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class Bottleneck(nn.Module):

    def __init__(self, in_planes, planes, stride=1, expansion=1):
        super(Bottleneck, self).__init__()
        self.expansion = expansion
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3,
                               stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, self.expansion *
                               planes, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(self.expansion*planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion*planes,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class ResNet(nn.Module):
    def __init__(self, block, num_blocks, in_ch, img_size = 128, num_classes=10, dropout=True):
        super(ResNet, self).__init__()
        self.in_planes = 16
        self.dropout_flag = dropout
        self.conv1 = nn.Conv2d(in_ch, 16, kernel_size=3,
                               stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(16)
        self.layer1 = self._make_layer(block, 16, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, 32, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, 32, num_blocks[2], stride=2)
        self.layer4 = self._make_layer(block, 64, num_blocks[3], stride=2)
        self.linear = nn.Linear(64 * int(img_size/2**5)**2 , num_classes)
        self.dropout1d = nn.Dropout(0.1)
        self.dropout2d = nn.Dropout2d(0.1)

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1]*(num_blocks-1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = F.max_pool2d(out, 4)
        if self.dropout_flag:
            out = self.dropout1d(out.view(out.size(0), -1))
        else:
            out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out


#==============================================================================
# 1D model
#==============================================================================
import torch
import torch.nn as nn
from einops import rearrange, repeat
import math

class ResNet_block(nn.Module):
    def __init__(self, n_inputs, n_outputs, kernel_size):
        super(ResNet_block, self).__init__()
        # conv1 (no downsampling)
        self.conv1 = nn.Conv1d(n_inputs, n_outputs, kernel_size[0],
                                           stride=1, padding=int((kernel_size[0]-1)/2), dilation=1)
        self.norm1 = nn.BatchNorm1d(n_outputs)
        self.act1 = nn.LeakyReLU()
        # conv2 (stride=2, downsample)
        self.conv2 = nn.Conv1d(n_outputs, n_outputs, kernel_size[1],
                               stride=2, padding=int((kernel_size[1]-1)/2), dilation=1)
        self.norm2 = nn.BatchNorm1d(n_outputs)
        self.act2 = nn.LeakyReLU()

        self.downsample = nn.AvgPool1d(2)
        self.conv3 = nn.Conv1d(n_inputs, n_outputs, kernel_size[1],
                               stride=1, padding=int((kernel_size[1]-1)/2), dilation=1)

    def forward(self, x):
        out = self.act1(self.norm1(self.conv1(x)))
        out = self.act2(self.norm2(self.conv2(out)))
        out = out + self.conv3(self.downsample(x))
        return out


class UpResNet_block(nn.Module):
    def __init__(self, n_inputs, n_outputs, kernel_size):
        super(UpResNet_block, self).__init__()
        # conv1 (no downsampling)
        self.conv1 = nn.Conv1d(n_inputs, n_outputs, kernel_size[0],
                               stride=1, padding=int((kernel_size[0] - 1) / 2), dilation=1)
        self.norm1 = nn.BatchNorm1d(n_outputs)
        self.act1 = nn.LeakyReLU()
        # conv2 (stride=2, upsample)
        self.conv2 = nn.ConvTranspose1d(n_outputs, n_outputs, kernel_size[1],
                               stride=2, padding=int((kernel_size[1] - 1) / 2), output_padding=1, dilation=1)
        self.norm2 = nn.BatchNorm1d(n_outputs)
        self.act2 = nn.LeakyReLU()

        self.upsample = nn.Upsample(scale_factor=2)
        self.conv3 = nn.Conv1d(n_inputs, n_outputs, kernel_size[1],
                               stride=1, padding=int((kernel_size[1] - 1) / 2), dilation=1)

    def forward(self, x):
        out = self.act1(self.norm1(self.conv1(x)))
        out = self.act2(self.norm2(self.conv2(out)))
        out = out + self.conv3(self.upsample(x))
        return out


class PositionalEncoding(nn.Module):

    def __init__(self, d_model, dropout=0.1, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0)
        self.register_buffer('pe', pe)

    def forward(self, x):
        x = x + self.pe[:, :x.size(1)]
        return self.dropout(x)


class PCGSingleLeadFeature(nn.Module):
    def __init__(self, n_inputs, res_channels, kernel_size,  heads, dropout, depth=4):
        super(PCGSingleLeadFeature, self).__init__()
        # conv unit
        self.num_res = len(res_channels)
        self.resnet = nn.ModuleList()
        for i in range(self.num_res):
            if i==0:
                self.resnet.append(ResNet_block(n_inputs, res_channels[0], kernel_size))
            else:
                self.resnet.append(ResNet_block(res_channels[i-1], res_channels[i], kernel_size))

        # 0-position class token
        dim = res_channels[-1]
        self.cls_token = nn.Parameter(torch.randn(1, 1, dim))

        # sin encoding
        self.sin_embedding = PositionalEncoding(d_model=dim)

        # attention unit
        encoder_layer = nn.TransformerEncoderLayer(d_model=dim, nhead=heads, dropout=dropout, batch_first=True)
        self.transformer_encoder = nn.TransformerEncoder(encoder_layer, num_layers=depth)

    def forward(self, x):
        for i in range(self.num_res):
            x = self.resnet[i](x)
        cls_tokens = repeat(self.cls_token, '1 n d -> b n d', b=x.shape[0])
        att_input = torch.cat((cls_tokens, x.permute(0, 2, 1)), dim=1)
        #att_input = x.permute(0, 2, 1)
        att_input = self.sin_embedding(att_input)
        att_out = self.transformer_encoder(att_input).permute(0, 2, 1)
        return att_out


class PCGSegmentation(nn.Module):
    def __init__(self, res_channels, kernel_size):
        super(PCGSegmentation, self).__init__()
        upres_channels = res_channels[::-1]
        self.num_res = len(upres_channels)
        self.resnet = nn.ModuleList()
        for i in range(self.num_res):
            if i != self.num_res-1:
                self.resnet.append(UpResNet_block(upres_channels[i], upres_channels[i+1], kernel_size))
            else:
                self.resnet.append(UpResNet_block(upres_channels[i], 5, kernel_size))
        #self.conv = nn.Conv1d(5, 5, kernel_size[0],
                                           #stride=1, padding=int((kernel_size[0] - 1) / 2), dilation=1)
        self.act = nn.Softmax(dim=1)

    def forward(self, x):
        for i in range(self.num_res):
            x = self.resnet[i](x)
        #x = self.conv(x)
        return self.act(x)


class PCGClassifier2D(nn.Module):
    def __init__(self, n_inputs, res_channels, num_classes):
        super(PCGClassifier2D, self).__init__()
        # conv unit
        self.num_res = len(res_channels)
        self.conv0 = nn.Conv2d(n_inputs, 8, kernel_size=3, stride=1, padding=1)
        self.bn0 = nn.BatchNorm2d(8)
        self.act0 = nn.ReLU()
        self.resnet = nn.ModuleList()
        for i in range(self.num_res):
            if i == 0:
                self.resnet.append(BasicBlock(8, res_channels[0], stride=2))
            else:
                self.resnet.append(BasicBlock(res_channels[i - 1], res_channels[i], stride=2))
        self.linear1 = nn.Linear(res_channels[-1] * 8**2, 512)
        self.act1 = nn.ReLU()
        self.linear2 = nn.Linear(512, 32)
        self.act2 = nn.ReLU()
        self.linear3 = nn.Linear(32, num_classes)
        self.dropout = nn.Dropout(0.1)

    def forward(self, x):
        x = self.act0(self.bn0(self.conv0(x)))
        for i in range(self.num_res):
            x = self.resnet[i](x)
        b = x.size(0)
        #x = x.view(b, -1)
        x = self.dropout(x.view(b,-1))
        x = self.linear3(self.act2(self.linear2(self.act1(self.linear1(x)))))
        return x


class PCGClassifierMix4(nn.Module):
    def __init__(self, n_inputs, n_outputs, kernel_size, num_classes):
        super(PCGClassifierMix4, self).__init__()
        self.res = ResNet_block(n_inputs, n_outputs, kernel_size)
        self.pool = nn.MaxPool1d(kernel_size=2)
        self.linear1 = nn.Linear(int(n_outputs*16), 32)
        self.act = nn.ReLU()
        self.linear2 = nn.Linear(32, num_classes)

    def forward(self, x):
        x = self.res(x)
        x = self.pool(x)
        b = x.size(0)
        x = x.view(b,-1).sum(dim=0, keepdim=True)
        x = self.linear2(self.act(self.linear1(x)))
        return x


class PCGModel(nn.Module):
    def __init__(self, num_classes, n_inputs, res_channels, kernel_size,  heads, dropout, depth=4):
        super(PCGModel, self).__init__()
        # feature
        self.feature_model = PCGSingleLeadFeature(n_inputs, res_channels, kernel_size,
                                                  heads, dropout, depth)
        dim = res_channels[-1]

        # classifier
        self.classifier = nn.Sequential(
            nn.LayerNorm(dim),
            nn.Linear(dim, 16),
            nn.ReLU(),
            nn.Linear(16, num_classes)
        )

        #self.classifier = PCGClassifier(dim, int(dim/2), [3, 3], num_classes)

        # segmentation
        self.segmentation = PCGSegmentation(res_channels, kernel_size)
        '''
            nn.Sequential(
            ResNet_block(dim, 4, kernel_size),
            nn.Upsample(scale_factor= 2**6),
            nn.Softmax(dim=1)
        )
        '''

    def forward(self, x):
        trans_features = self.feature_model(x)
        pred_seg = self.segmentation(trans_features[:,:,1:])
        class_out = self.classifier(trans_features[:,:,0])
        return pred_seg, class_out



# Paper: Deep Convolutional Neural Networks for Heart Sound Segmentation
# Unet segmentation + Max temporal modelling
class UnetSegmentation1D(nn.Module):
    def __init__(self):
        super(UnetSegmentation1D, self).__init__()
