import numpy as np
from scipy.signal import decimate
import scipy
from numpy.fft import rfft, irfft, rfftfreq
import os
import pandas as pd
from helper_code import *
from sklearn.model_selection import StratifiedKFold
#import librosa
import torch.nn as nn
import torch.nn.functional as F
import torch

def get_murmur_locations(data):
    num_locations = get_num_locations(data)
    locations = list()
    for i, l in enumerate(data.split('\n')):
        entries = l.split(' ')
        if i==0:
            pass
        elif 1<=i<=num_locations:
            locations.append(entries[0])
        else:
            break
    return locations


def get_systolic_murmur(data):
    sm = None
    for l in data.split('\n'):
        if l.startswith('#Systolic murmur timing:'):
            try:
                sm = l.split(': ')[1].strip()
            except:
                pass
    if sm is not None:
        if sm=='nan':
            sm = 0
        else:
            sm = 1
    else:
        sm = 0
    return sm


def get_diastolic_murmur(data):
    dm = None
    for l in data.split('\n'):
        if l.startswith('#Diastolic murmur timing:'):
            try:
                dm = l.split(': ')[1].strip()
            except:
                pass
    if dm is not None:
        if dm=='nan':
            dm = 0
        else:
            dm = 1
    else:
        dm = 0
    return dm


# loss functions
# PyTorch
#https://www.kaggle.com/code/bigironsphere/loss-function-library-keras-pytorch/notebook
class DiceBCELoss(nn.Module):
    def __init__(self, weight=None, size_average=True):
        super(DiceBCELoss, self).__init__()

    def forward(self, inputs, targets, smooth=1):
        # flatten label and prediction tensors
        inputs = inputs.view(-1)
        targets = targets.view(-1)

        intersection = (inputs * targets).sum()
        dice_loss = 1 - (2. * intersection + smooth) / (inputs.sum() + targets.sum() + smooth)
        BCE = F.binary_cross_entropy(inputs, targets, reduction='mean')
        Dice_BCE = BCE + dice_loss

        return Dice_BCE


class DiceLoss(nn.Module):
    def __init__(self, weight=None, size_average=True):
        super(DiceLoss, self).__init__()

    def forward(self, inputs, targets, smooth=1):
        # flatten label and prediction tensors
        intersection = torch.sum(inputs * targets, dim=2, keepdim=True)
        input_pow = torch.sum(inputs.pow(2), dim=2, keepdim=True)
        target_pow = torch.sum(targets.pow(2), dim=2, keepdim=True)
        out = 1 - ((2. * intersection + smooth) /
                   (input_pow + target_pow + smooth))
        return out.mean()


class IoULoss(nn.Module):
    def __init__(self, weight=None, size_average=True):
        super(IoULoss, self).__init__()

    def forward(self, inputs, targets, smooth=1):
        # comment out if your model contains a sigmoid or equivalent activation layer
        inputs = F.sigmoid(inputs)

        # flatten label and prediction tensors
        inputs = inputs.view(-1)
        targets = targets.view(-1)

        # intersection is equivalent to True Positive count
        # union is the mutually inclusive area of all labels & predictions
        intersection = (inputs * targets).sum()
        total = (inputs + targets).sum()
        union = total - intersection

        IoU = (intersection + smooth) / (union + smooth)

        return 1 - IoU

def load_all_channel_records(data_folder, data, get_frequencies=False):
    record_dict = {}
    recording_locations = ['AV', 'MV', 'PV', 'TV']
    for r in recording_locations:
        record_dict[r] = None

    num_locations = get_num_locations(data)
    recording_information = data.split('\n')[1:num_locations + 1]

    frequencies = list()
    for i in range(num_locations):
        entries = recording_information[i].split(' ')
        recording_file = entries[2]
        filename = os.path.join(data_folder, recording_file)
        recording, frequency = load_wav_file(filename)
        if record_dict[entries[0]] is None:
            record_dict[entries[0]] = recording
        frequencies.append(frequency)

    if get_frequencies:
        return record_dict, frequencies
    else:
        return record_dict


def create_fold_info(info_df):
    X = info_df['ID'].values
    y = info_df['Murmur'].values
    skf = StratifiedKFold(n_splits=10)
    folds = np.zeros(len(X))
    i = 0
    for train_index, test_index in skf.split(X, y):
        folds[test_index] = int(i + 1)
        i += 1
    info_df['Fold'] = folds
    return info_df


def create_single_lead_csv(files, data_folder, model_folder, save=False):
    # get patient ID, channel data, channel label, create list for 2022 dataset
    all_data_info = []
    murmur_locations_info = []
    outcome_locations_info = []
    for filepath in files:
        header = load_patient_data(filepath)
        # get id,fre,num_locations
        idp = get_patient_id(header)
        fre = get_frequency(header)
        num_locations = get_num_locations(header)
        murmur_locations = None
        murmur_label = 'Unknown'
        outcome_label = 'Abnormal'
        for l in header.split('\n'):
            if l.startswith('#Murmur:'):
                try:
                    murmur_label = l.split(':')[1].strip()
                except:
                    pass
            if '#Murmur locations' in l:
                try:
                    murmur_locations = l.split(':')[1].strip().split('+')
                except:
                    pass
            if l.startswith('#Outcome:'):
                try:
                    outcome_label = l.split(':')[1].strip()
                except:
                    pass
        all_data_info.append([idp, murmur_label, outcome_label])
        murmur_dict = {}
        outcome_dict = {}
        recording_information = header.split('\n')[1:num_locations + 1]
        for i in range(num_locations):
            entries = recording_information[i].split(' ')
            location = entries[0]
            recording_file = entries[2]
            mask_file = entries[3]
            # present:0, unknown:1, absent:2
            values = [idp, fre, os.path.join(data_folder, recording_file), os.path.join(data_folder, mask_file)]
            if murmur_label.casefold() == 'Unknown'.casefold():
                values.append(1)
            elif murmur_label.casefold() == 'Absent'.casefold():
                values.append(2)
            elif murmur_label.casefold() == 'Present'.casefold():
                if murmur_locations is not None:
                    if location in murmur_locations:
                        values.append(0)
                    else:
                        values.append(1)
                else:
                    values.append(1)
            else:
                print('Invalid class name')

            murmur_dict[location] = values
            # abnormal:0, normal:1
            values2 = [idp, fre, os.path.join(data_folder, recording_file), os.path.join(data_folder, mask_file)]
            if outcome_label.casefold() == 'Abnormal'.casefold():
                values2.append(0)
            else:
                values2.append(1)
            outcome_dict[location] = values2
        for key in murmur_dict.keys():
            cur_arr = murmur_dict[key]
            cur_arr.append(key)
            murmur_locations_info.append(cur_arr)
        for key in outcome_dict.keys():
            cur_arr = outcome_dict[key]
            cur_arr.append(key)
            outcome_locations_info.append(cur_arr)

    murmur_locations_info = np.array(murmur_locations_info)
    outcome_locations_info = np.array(outcome_locations_info)
    murmur_locations_df = pd.DataFrame(murmur_locations_info,
                                       columns=['ID', 'Frequency', 'FileName', 'MaskFile', 'Label', 'Location'])
    outcome_locations_df = pd.DataFrame(outcome_locations_info,
                                       columns=['ID', 'Frequency', 'FileName', 'MaskFile', 'Label', 'Location'])

    all_data_info = np.array(all_data_info)
    all_data_df = pd.DataFrame(all_data_info, columns=['ID', 'Murmur', 'Outcome'])
    # change outcome to category numbers
    outcome = all_data_df['Outcome'].values
    idx = np.where(outcome=='Normal')
    outcome_ls = np.zeros(len(outcome))
    outcome_ls[idx] = 1
    all_data_df['Outcome'] = outcome_ls
    all_data_df = create_fold_info(all_data_df)
    murmur_out_df = pd.merge(murmur_locations_df, all_data_df, how='left', on='ID')
    outcome_out_df = pd.merge(outcome_locations_df, all_data_df, how='left', on='ID')
    if save:
        murmur_out_df.to_csv(os.path.join(model_folder, 'murmur_locations_files.csv'), index=False)
        outcome_out_df.to_csv(os.path.join(model_folder, 'outcome_locations_files.csv'), index=False)
    return all_data_df, murmur_out_df, outcome_out_df


def get_train_valid_df(df, fold=1):
    valid_df = df[df['Fold']==fold]
    train_df = df[df['Fold']!=fold]
    return train_df, valid_df


#====================================================================
# Preprocessing
#====================================================================
def downsample(signal, factor=4):
    record = decimate(signal, factor, ftype='fir')
    return record


def filtering(signal, low=25, high=400, order=4, sampling_rate=1000):
    sos = scipy.signal.butter(order, [low, high], btype='bandpass', output="sos", fs=sampling_rate)
    filtered = scipy.signal.sosfiltfilt(sos, signal)
    return filtered


def remove_segment_max(segment):
    pos = np.argmax(np.abs(segment))
    # get start pos (zero_crossing point)
    start = 0
    for i in range(pos,1,-1):
        if segment[i] * segment[i-1] <= 0:
            start = i-1
            break
    # get end pos
    end = len(segment)
    for i in range(pos, len(segment)-1):
        if segment[i] * segment[i+1] <= 0:
            end = i+1
            break
    segment[start:end] = 0
    return segment


def remove_spikes(signal, terminal=500, sampling=1000):
    w_pts = int(terminal/1000 * sampling)
    splits = np.arange(0, len(signal), w_pts)
    if splits[-1]!=len(signal):
        splits=np.append(splits, len(signal))
    num_m = np.inf
    while True:
        maas = []
        for s,e in zip(splits[:-1], splits[1:]):
            sigslice = signal[s:e]
            maa = np.max(np.abs(sigslice))
            maas.append(maa)
        maas = np.array(maas)
        median = np.median(maas)
        pos = np.where(maas>3*median)[0]
        num_m = len(pos)
        if num_m>0:
            # get largest point window segment
            m = pos[np.argmax(maas[pos])]
            s = splits[m]
            e = splits[m+1]
            segment = signal[s:e]
            segment = remove_segment_max(segment)
            signal[s:e] = segment
        else:
            break
    return signal


def normalize_signal(signal):
    # scale to [-1,1]
    sort_sig = np.sort(signal)
    low = np.median(sort_sig[:int(0.001 * len(signal))])
    high = np.median(sort_sig[-int(0.001 * len(signal)):])
    v100 = max(abs(low), abs(high))
    # v100 = np.max(np.abs(signal))
    if v100==0:
        signal = signal
    else:
        signal = signal / v100
    return signal


def decompose_into_n_signals(srs, fre_split = [25,45,85,200,400], sampling_rate=1000):
    signal_length = len(srs)
    fourier = rfft(srs)
    frequencies = rfftfreq(srs.size, d=1./sampling_rate)
    decomp = []
    for f1, f2 in zip(fre_split[:-1], fre_split[1:]):
        cur_fft = fourier.copy()
        cur_fft[(frequencies>f2)] = 0
        cur_fft[(frequencies<f1)] = 0
        decomp.append(irfft(cur_fft))
    out_decomp = []
    for sig in decomp:
        if len(sig) < signal_length:
            out_decomp.append(np.pad(sig, (0, signal_length - len(sig))))
        elif len(sig) > signal_length:
            out_decomp.append(sig[:signal_length])
        else:
            out_decomp.append(sig)
    return out_decomp


def create_2d_sepctrogram(signal, hop_length=64, n_fft=512, sample_rate=4000):
    # STFT -> spectrogram
    # perform stft
    #stft_normal = librosa.stft(signal.astype(float), n_fft=n_fft, hop_length=hop_length)

    # calculate abs values on complex numbers to get magnitude
    #spectrogram = np.abs(stft_normal)
    #log_spectrogram = librosa.amplitude_to_db(spectrogram)
    #fre_cols = librosa.fft_frequencies(sr=sample_rate, n_fft=n_fft)
    #end_idx = np.where(fre_cols==500)[0][0]
    #out_img = log_spectrogram[:end_idx]
    return None#out_img


from scipy.signal import periodogram, welch
from scipy.signal import spectrogram
def get_psd_features(record):
    # downsample
    drecordings = downsample(record, factor=4)
    # filtering
    frecordings = filtering(drecordings, low=25, high=400, order=4, sampling_rate=1000)
    # remove spikes
    srecordings = remove_spikes(frecordings, terminal=500, sampling=1000)
    # normalize
    nrecordings = normalize_signal(srecordings)
    # welch psd 
    x = nrecordings
    fs = 1000
    freqs, psd = welch(x, fs, nperseg=512)
    psd_total = []
    frac_fre = freqs[1]-freqs[0]
    interval_list = np.arange(20,420,20)
    for i in range(len(interval_list)-1):
        pos = set(list(np.where(freqs>interval_list[i])[0])).intersection(set(list(np.where(freqs<interval_list[i+1])[0])))
        psd_total.append(psd[list(pos)])
    psd_totalf = []
    for p in psd_total:
        psd_totalf.append([np.mean(p), np.min(p), np.max(p), np.std(p), np.sum(p)*frac_fre])
    psd_totalf = np.array(psd_totalf).flatten()
    # spectrogram 
    x = nrecordings
    fs = 1000
    f, t, Sxx = spectrogram(x, fs, window=('gaussian',75))
    # cut = 10s , 50pts 
    num_pts = Sxx.shape[1]
    if num_pts < 50:
        Sxx = np.pad(Sxx, ((0,0), (0, 50-num_pts)))
    else:
        s = np.random.randint(0, num_pts - 50 + 1, 1)[0]
        Sxx = Sxx[:, s:s+50]
    psd_pos = []
    interval_list = [25,45,85,200,400]
    for i in range(len(interval_list)-1):
        pos = set(list(np.where(f>interval_list[i])[0])).intersection(set(list(np.where(f<interval_list[i+1])[0])))
        psd_pos.append(list(pos))
    psd_time = [Sxx[pos, :] for pos in psd_pos]
    psd_timef = []
    for p in psd_time:
        psd_timef.append(np.mean(p, axis=0))
        psd_timef.append(np.std(p, axis=0))
        psd_timef.append(np.max(p, axis=0))
        psd_timef.append(np.min(p, axis=0))
    psd_timef = np.array(psd_timef).flatten()
    return np.concatenate([psd_totalf, psd_timef])
