import sys
from helper_code import is_integer
from dataset import *
from model import *
import torch.nn as nn
from time import time
from myutils import *
import torch
import matplotlib.pyplot as plt
from sklearn.metrics import classification_report

def train_outcome_classifier(model_folder, locations_info, fold, verbose):
    # ============================================================================================================
    # Present Unknown Absent classifier
    # ============================================================================================================
    classes = ['Abnormal', 'Normal']
    num_classes = len(classes)

    # Extract the features and labels.
    if verbose >= 1:
        print('Set dataset.')
    # ============================================================================================================
    # ============================================================================================================
    # create single dataset
    batch = 128
    train_df, valid_df = get_train_valid_df(locations_info, fold=fold)
    train_ = MurmurSingleChannelHeartBeatDataset2D(train_df, img_size=64, hop_length=32, n_fft=256, dataset=2016, prob=0.8)
    valid_ = MurmurSingleChannelHeartBeatDataset2D(valid_df, img_size=64, hop_length=32, n_fft=256, dataset=2016, prob=0.8)
    print('Train', 'Valid')
    print(len(train_), len(valid_))
    train_sampler = torch.utils.data.sampler.WeightedRandomSampler(torch.from_numpy(train_.weights), len(train_.weights))
    train_loader = torch.utils.data.DataLoader(train_,
                                               batch_size=batch, sampler = train_sampler)
    valid_loader = torch.utils.data.DataLoader(valid_,
                                               batch_size=1, shuffle=False,
                                               drop_last=False)
    # set model
    model = PCGClassifier2D(1, res_channels=[8, 16, 32], num_classes=2)
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    model.to(device)

    # set training parameters
    lr = 1e-4
    lr_step = True
    optimizer = torch.optim.Adam(model.parameters(), lr=lr)
    if lr_step:
        scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[5, 10, 20, 40], gamma=0.5)
    loss_class = nn.CrossEntropyLoss()

    # begin training
    if verbose >= 1:
        print('Begin training.')
    epochs = 50
    cur_valid_loss = np.inf
    for epoch in range(1, epochs + 1):
        # ===================================================
        # Train
        # ===================================================
        model.train()
        epoch_loss = []
        epoch_step_time = []
        for i, sample in enumerate(train_loader, 1):
            step_start_time = time()
            signal, label = sample
            signal = signal.to(device, dtype=torch.float)
            label = label.to(device, dtype=torch.long)
            predicts_class = model(signal)
            total_loss = loss_class(predicts_class, label)
            optimizer.zero_grad()
            total_loss.backward()  # (retain_graph=True)
            optimizer.step()
            epoch_loss.append(total_loss.item())
            if lr_step:
                scheduler.step()
            epoch_step_time.append(time() - step_start_time)
        epoch_info = 'Training Epoch %d/%d' % (epoch, epochs)
        time_info = '%.4f sec/step' % np.mean(epoch_step_time)
        loss_info = 'loss: %.4e ' % (np.mean(epoch_loss))
        print(' - '.join((epoch_info, time_info, loss_info)), flush=True)

        # ===================================================
        # validation
        # ===================================================
        with torch.no_grad():
            model.eval()
            valid_epoch_loss = []
            valid_gt = []
            valid_pred = []
            epoch_step_time = []
            for i, sample in enumerate(valid_loader, 1):
                step_start_time = time()
                signal, label = sample
                signal = signal.to(device, dtype=torch.float)
                label = label.to(device, dtype=torch.long)
                predicts_class = model(signal)
                preds = np.argmax(predicts_class.detach().cpu().numpy(), axis=1)
                gts = label.detach().cpu().numpy()
                valid_gt.append(gts)
                valid_pred.append(preds)
                total_loss = loss_class(predicts_class, label)
                valid_epoch_loss.append(total_loss.item())
                if lr_step:
                    scheduler.step()
                epoch_step_time.append(time() - step_start_time)
            valid_gt = np.array(valid_gt)
            valid_pred = np.array(valid_pred)
            print(classification_report(valid_gt, valid_pred))
            epoch_info = 'Validation Epoch %d/%d' % (epoch, epochs)
            time_info = '%.4f sec/step' % np.mean(epoch_step_time)
            valid_loss_info = 'loss: %.4e' % (np.mean(valid_epoch_loss))
            print(' - '.join((epoch_info, time_info, valid_loss_info)), flush=True)
            with open(os.path.join(model_folder, 'outcome_pretrain_validation_fold' + str(fold) + '.csv'), 'a') as f:
                f.write(str(epoch) + ',' + '%.4e' % (np.mean(epoch_loss)) + ',' + '%.4e' % (
                    np.mean(valid_epoch_loss)) + '\n')
        if np.mean(valid_epoch_loss) < cur_valid_loss:
            cur_valid_loss = np.mean(valid_epoch_loss)
            # save model
            torch.save(model.state_dict(), os.path.join(model_folder, 'outcome_pretrain_best_model.pt'))
    if verbose >= 1:
        print('Done.')


def train_PCG_segmentation(model_folder, locations_info, fold, verbose):
    # ============================================================================================================
    # Present Unknown Absent classifier
    # ============================================================================================================
    classes = ['Abnormal', 'Normal']
    num_classes = len(classes)

    # Extract the features and labels.
    if verbose >= 1:
        print('Set dataset.')
    # ============================================================================================================
    # ============================================================================================================
    # create single dataset
    batch = 128
    train_df, valid_df = get_train_valid_df(locations_info, fold=fold)
    train_ = MurmurSingleChannelDataset1D(train_df, resize_scale=16, signal_length=8192, decompose=True, dataset=2016)
    valid_ = MurmurSingleChannelDataset1D(valid_df, resize_scale=16, signal_length=8192, decompose=True, dataset=2016)
    print('Train', 'Valid')
    print(len(train_), len(valid_))
    train_sampler = torch.utils.data.sampler.WeightedRandomSampler(torch.from_numpy(train_.weights), len(train_.weights))
    train_loader = torch.utils.data.DataLoader(train_,
                                               batch_size=batch, sampler = train_sampler)
    valid_loader = torch.utils.data.DataLoader(valid_,
                                               batch_size=batch, shuffle=False,
                                               drop_last=False)
    # set model
    model = PCGModel(2, n_inputs=5, res_channels=[8, 16, 32, 64], kernel_size=[3, 3],
                     heads=4, dropout=0.2, depth=1)
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    model.to(device)
    sigmoid_layer = nn.Sigmoid()
    sigmoid_layer.to(device)

    # set training parameters
    lr = 1e-3
    lr_step = False
    optimizer = torch.optim.Adam(model.parameters(), lr=lr)
    if lr_step:
        scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[10, 30, 50, 80, 120, 150], gamma=0.5)
    loss_class = nn.CrossEntropyLoss()
    loss_seg = DiceLoss()

    # begin training
    if verbose >= 1:
        print('Begin training.')
    epochs = 100
    cur_valid_loss = np.inf
    for epoch in range(1, epochs + 1):
        # ===================================================
        # Train
        # ===================================================
        model.train()
        epoch_loss = []
        epoch_segloss = []
        epoch_classloss = []
        epoch_step_time = []
        for i, sample in enumerate(train_loader, 1):
            step_start_time = time()
            signal, mask, mask4, label = sample
            signal = signal.to(device, dtype=torch.float)
            mask4 = mask4.to(device, dtype=torch.float)
            mask = mask.to(device, dtype=torch.long)
            label = label.to(device, dtype=torch.long)
            predicts_seg, predicts_class = model(signal)
            ce_loss = torch.tensor(0)#loss_class(predicts_class, label)
            dice_loss = loss_seg(predicts_seg, mask4)
            total_loss = dice_loss + 5*ce_loss
            optimizer.zero_grad()
            total_loss.backward()  # (retain_graph=True)
            optimizer.step()
            epoch_loss.append(total_loss.item())
            epoch_classloss.append(ce_loss.item())
            epoch_segloss.append(dice_loss.item())
            if lr_step:
                scheduler.step()
            epoch_step_time.append(time() - step_start_time)
        epoch_info = 'Training Epoch %d/%d' % (epoch, epochs)
        time_info = '%.4f sec/step' % np.mean(epoch_step_time)
        loss_info = 'loss: %.4e (Dice %.4e, CE %.4e) ' % (np.mean(epoch_loss), np.mean(epoch_segloss),
                                                          np.mean(epoch_classloss))
        print(' - '.join((epoch_info, time_info, loss_info)), flush=True)

        # ===================================================
        # validation
        # ===================================================
        with torch.no_grad():
            model.eval()
            valid_epoch_loss = []
            epoch_segloss = []
            epoch_classloss = []
            epoch_step_time = []
            for i, sample in enumerate(valid_loader, 1):
                step_start_time = time()
                signal, mask, mask4, label = sample
                signal = signal.to(device, dtype=torch.float)
                mask4 = mask4.to(device, dtype=torch.float)
                mask = mask.to(device, dtype=torch.long)
                label = label.to(device, dtype=torch.long)
                predicts_seg, predicts_class = model(signal)
                ce_loss = torch.tensor(0)#loss_class(predicts_class, label)
                dice_loss = loss_seg(predicts_seg, mask4)
                total_loss = dice_loss + 5*ce_loss
                valid_epoch_loss.append(total_loss.item())
                epoch_classloss.append(ce_loss.item())
                epoch_segloss.append(dice_loss.item())
                if lr_step:
                    scheduler.step()
                epoch_step_time.append(time() - step_start_time)
            epoch_info = 'Validation Epoch %d/%d' % (epoch, epochs)
            time_info = '%.4f sec/step' % np.mean(epoch_step_time)
            valid_loss_info = 'loss: %.4e (Dice %.4e, CE %.4e) ' % (np.mean(valid_epoch_loss), np.mean(epoch_segloss),
                                                              np.mean(epoch_classloss))
            print(' - '.join((epoch_info, time_info, valid_loss_info)), flush=True)
            with open(os.path.join(model_folder, 'segmentation_pretrain_validation_fold' + str(fold) + '.csv'), 'a') as f:
                f.write(str(epoch) + ',' + '%.4e' % (np.mean(epoch_loss)) + ',' + '%.4e' % (
                    np.mean(valid_epoch_loss)) + '\n')
        if np.mean(valid_epoch_loss) < cur_valid_loss:
            cur_valid_loss = np.mean(valid_epoch_loss)
            # save model
            torch.save(model.state_dict(), os.path.join(model_folder, 'segmentation_pretrain_best_model.pt'))
    if verbose >= 1:
        print('Done.')


def evaluate_outcome_classifier(model_folder, locations_info, fold, verbose):
    # ============================================================================================================
    # Present Unknown Absent classifier
    # ============================================================================================================
    classes = ['Abnormal', 'Normal']
    num_classes = len(classes)

    # Extract the features and labels.
    if verbose >= 1:
        print('Set dataset.')
    # ============================================================================================================
    # ============================================================================================================
    # create single dataset
    batch = 128
    train_df, valid_df = get_train_valid_df(locations_info, fold=fold)
    train_ = MurmurSingleChannelHeartBeatDataset2D(train_df, img_size=64, hop_length=32, n_fft=256, dataset=2016,
                                                   prob=0.8)
    valid_ = MurmurSingleChannelHeartBeatDataset2D(valid_df, img_size=64, hop_length=32, n_fft=256, dataset=2016,
                                                   prob=0.8)
    print('Train', 'Valid')
    print(len(train_), len(valid_))
    train_sampler = torch.utils.data.sampler.WeightedRandomSampler(torch.from_numpy(train_.weights),
                                                                   len(train_.weights))
    train_loader = torch.utils.data.DataLoader(train_,
                                               batch_size=batch, sampler=train_sampler)
    valid_loader = torch.utils.data.DataLoader(valid_,
                                               batch_size=batch, shuffle=False,
                                               drop_last=False)
    # set model
    model = PCGClassifier2D(1, res_channels=[8, 16, 32], num_classes=2)
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    model.to(device)
    if device == 'cpu':
        state_dict = torch.load(os.path.join(model_folder, 'outcome_pretrain_best_model.pt'), map_location='cpu')
    else:
        state_dict = torch.load(os.path.join(model_folder, 'outcome_pretrain_best_model.pt'))
    model.load_state_dict(state_dict)
    sigmoid_layer = nn.Sigmoid()
    sigmoid_layer.to(device)

    with torch.no_grad():
        model.eval()
        for i, sample in enumerate(valid_loader, 1):
            signal, label = sample
            signal = signal.to(device, dtype=torch.float)
            label = label.to(device, dtype=torch.long)
            predicts_class = model(signal)
            pred_class = torch.argmax(sigmoid_layer(predicts_class), dim=1)
            print(pred_class, label)
            break



if __name__ == '__main__':
    # Parse the arguments.
    if not (len(sys.argv) == 3 or len(sys.argv) == 4):
        raise Exception('Include the data and model folders as arguments, e.g., python train_model.py data model.')

    # Define the data and model foldes.
    data_folder = sys.argv[1]
    model_folder = sys.argv[2]

    # Change the level of verbosity; helpful for debugging.
    if len(sys.argv)==4 and is_integer(sys.argv[3]):
        verbose = int(sys.argv[3])
    else:
        verbose = 1

    # create file csv
    names = ['a', 'b', 'c', 'd', 'e', 'f']
    final_csv = pd.DataFrame()
    for name in names:
        cur_folder = os.path.join(data_folder, 'training-' + name)
        cur_csv = pd.read_csv(cur_folder + '/REFERENCE.csv', header=None)
        cur_csv.columns = ['ID', 'Label']
        cur_csv['Label'] = 1 - (cur_csv['Label'].values + 1) / 2
        idx = cur_csv['ID'].values
        cur_csv['FileName'] = data_folder + '/training-' + name + '/' + idx + '.wav'
        cur_csv['MaskFile'] = data_folder[:-9] + '/training_annotations/hand_corrected/training-' + name +\
                              '_StateAns/' + idx + '_StateAns.mat'
        final_csv = pd.concat([cur_csv, final_csv])
    valid_csv = pd.read_csv(data_folder[:-9] + '/validation/REFERENCE.csv', header=None)
    valid_csv.columns = ['ID', 'Other']
    valid_csv['Fold'] = 1
    final_csv = pd.merge(final_csv, valid_csv, on='ID', how='left')
    final_csv['Fold'] = final_csv['Fold'].fillna(0)

    exist_annots = []
    for i in names:
        folder = data_folder[:-9] + '/training_annotations/hand_corrected/training-' + i + '_StateAns'
        files = os.listdir(folder)
        fullfiles = [os.path.join(folder, f) for f in files]
        exist_annots.append(fullfiles)
    exist_df = pd.DataFrame({'MaskFile': np.concatenate(exist_annots)})
    final_csv = pd.merge(exist_df, final_csv, on='MaskFile', how='inner')

    train_outcome_classifier(model_folder, final_csv, 1, verbose)
    #evaluate_outcome_classifier(model_folder, final_csv, 1, verbose)